@@ -53,11 +53,6 @@ extern CFRunLoopSourceRef gUimRun;
 extern CFSocketContext gSockContext;
 
 
-//#define DEBUG_HELPER  1
-//#define DEBUG_PREEDIT  1
-//#define DEBUG_CANDIDATES  1
-
-
 static void
 AddPreeditSegment(MUIMSessionHandle inHandle, int inAttr,
                   const char *inStr);
@@ -73,7 +68,7 @@ ParseHelperString(const char *str);
 void
 UIMCommitString(void *ptr, const char *str)
 {
-#ifdef DEBUG_PREEDIT
+#if DEBUG_PREEDIT
   DEBUG_PRINT("UIMCommitString() len=%d\n", strlen(str));
 #endif
 
@@ -107,7 +102,7 @@ UIMPreeditClear(void *ptr)
   MUIMSessionHandle handle = gActiveSession;
   UInt32 i;
 
-#ifdef DEBUG_PREEDIT
+#if DEBUG_PREEDIT
   DEBUG_PRINT("UIMPreeditClear()\n");
 #endif
 
@@ -125,43 +120,32 @@ UIMPreeditPushback(void *ptr, int attr, const char *str)
   //MUIMSessionHandle handle = (MUIMSessionHandle) ptr;
   MUIMSessionHandle handle = gActiveSession;
 
-#ifdef DEBUG
+#if DEBUG_PREEDIT
   {
-    char *tmpstr = "";
-    char attrstr[50];
+    char attrstr[50] = { ' ', '\0' };
 
-    if (attr & UPreeditAttr_None) {
-      tmpstr = "None";
-      sprintf(attrstr, "%s", tmpstr);
-    }
-    if (attr & UPreeditAttr_UnderLine) {
-      tmpstr = "UnderLine";
-      sprintf(attrstr, "%s %s", attrstr, tmpstr);
-    }
-    if (attr & UPreeditAttr_Reverse) {
-      tmpstr = "Reverse";
-      sprintf(attrstr, "%s %s", attrstr, tmpstr);
-    }
-    if (attr & UPreeditAttr_Cursor) {
-      tmpstr = "Cursor";
-      sprintf(attrstr, "%s %s", attrstr, tmpstr);
-    }
-    if (attr & UPreeditAttr_Separator) {
-      tmpstr = "Separator";
-      sprintf(attrstr, "%s %s", attrstr, tmpstr);
-    }
+    if (attr & UPreeditAttr_None)
+      strcpy(attrstr, "None");
+    if (attr & UPreeditAttr_UnderLine)
+      strcat(attrstr, " UnderLine");
+    if (attr & UPreeditAttr_Reverse)
+      strcat(attrstr, " Reverse");
+    if (attr & UPreeditAttr_Cursor)
+      strcat(attrstr, " Cursor");
+    if (attr & UPreeditAttr_Separator)
+      strcat(attrstr, " Separator");
 
-#ifdef DEBUG_PREEDIT
-    DEBUG_PRINT("UIMPreeditPushback() fSegmentCount=%lu attr=%s len=%lu\n",
+    DEBUG_PRINT("UIMPreeditPushback() fSegmentCount=%lu attr=%s len=%lu str='%s'\n",
                 (*handle)->fSegmentCount,
-                attrstr, strlen(str));
-#endif
+                attrstr, strlen(str), str);
   }
 #endif
 
+  /*
   if (!strcmp(str, "")
       && !(attr & (UPreeditAttr_Cursor | UPreeditAttr_Separator | UPreeditAttr_UnderLine)))
     return;
+  */
 
   AddPreeditSegment(handle, attr, str);
 }
@@ -172,7 +156,7 @@ UIMPreeditUpdate(void *ptr)
   //MUIMSessionHandle handle = (MUIMSessionHandle) ptr;
   MUIMSessionHandle handle = gActiveSession;
 
-#ifdef DEBUG_PREEDIT
+#if DEBUG_PREEDIT
   DEBUG_PRINT("UIMPreeditUpdate()\n");
 #endif
 
@@ -188,7 +172,7 @@ UIMCandAcivate(void *inPtr, int inNR, int inLimit)
   //MUIMSessionHandle handle = (MUIMSessionHandle) inPtr;
   MUIMSessionHandle handle = gActiveSession;
 
-#ifdef DEBUG_CANDIDATES
+#if DEBUG_CANDIDATES
   DEBUG_PRINT("UIMCandAcivate() inNR=%d inLimit=%d\n", inNR, inLimit);
 #endif
 
@@ -212,7 +196,7 @@ UIMCandSelect(void *inPtr, int inIndex)
   //MUIMSessionHandle handle = (MUIMSessionHandle) inPtr;
   MUIMSessionHandle handle = gActiveSession;
 
-#ifdef DEBUG_CANDIDATES
+#if DEBUG_CANDIDATES
   DEBUG_PRINT("UIMCandSelect() inIndex=%d\n", inIndex);
 #endif
 
@@ -244,7 +228,7 @@ UIMCandShiftPage(void *inPtr, int inDirection)
       (*handle)->fCandidateIndex -= (*handle)->fDisplayLimit;
   }
 
-#ifdef DEBUG_CANDIDATES
+#if DEBUG_CANDIDATES
   DEBUG_PRINT("UIMCandShiftPage() inDirection=%d fCandidateIndex=%ld\n",
               inDirection, (*handle)->fCandidateIndex);
 #endif
@@ -262,7 +246,7 @@ UIMCandDeactivate(void *inPtr)
   //MUIMSessionHandle handle = (MUIMSessionHandle) inPtr;
   MUIMSessionHandle handle = gActiveSession;
 
-#ifdef DEBUG_CANDIDATES
+#if DEBUG_CANDIDATES
   DEBUG_PRINT("UIMCandDeactivate()\n");
 #endif
 
@@ -274,7 +258,7 @@ UIMUpdatePropList(void *inPtr, const char *inStr)
 {
   char *tmp;
 
-#ifdef DEBUG_HELPER
+#if DEBUG_HELPER
   DEBUG_PRINT("UIMUpdatePropList() gActiveSession=%p inStr='%s'\n",
               gActiveSession, inStr);
 #endif
@@ -300,7 +284,7 @@ UIMUpdatePropLabel(void *inPtr, const char *inStr)
 {
   char *tmp;
 
-#ifdef DEBUG_HELPER
+#if DEBUG_HELPER
   DEBUG_PRINT("UIMUpdatePropLabel() gActiveSession=%p inStr='%s'\n",
               gActiveSession, inStr);
 #endif
@@ -327,7 +311,7 @@ UIMCheckHelper()
   if (gUimFD < 0) {
     gUimFD = uim_helper_init_client_fd(UIMHelperDisconnect);
 
-#ifdef DEBUG_HELPER
+#if DEBUG_HELPER
     DEBUG_PRINT("UIMCheckHelper() gUimFD=%d(%p)\n", gUimFD, &gUimFD);
 #endif
 
@@ -354,7 +338,7 @@ UIMCheckHelper()
           return;
         }
         CFRunLoopAddSource(CFRunLoopGetCurrent(), gUimRun, kCFRunLoopDefaultMode);
-#ifdef DEBUG_HELPER
+#if DEBUG_HELPER
         DEBUG_PRINT("UIMCheckHelper() CFRunLoopGetCurrent()\n");
 #endif
       }
@@ -368,7 +352,7 @@ HelperRead(CFSocketRef sock, CFSocketCallBackType callbackType,
 {
   char *tmp;
 
-#ifdef DEBUG_HELPER
+#if DEBUG_HELPER
   DEBUG_PRINT("HelperRead()\n");
 #endif
 
@@ -384,7 +368,7 @@ ParseHelperString(const char *str)
 {
   UInt32 i;
 
-#ifdef DEBUG_HELPER
+#if DEBUG_HELPER
   DEBUG_PRINT("ParseHelperString() str='%s'\n", str);
 #endif
 
@@ -410,7 +394,7 @@ ParseHelperString(const char *str)
           (first = CFArrayGetValueAtIndex(array, 0))) {
         CFStringRef second = CFArrayGetValueAtIndex(array, 1);
         if (second) {
-#ifdef DEBUG_HELPER
+#if DEBUG_HELPER
           DEBUG_PRINT("ParseHelperString() first='%s' second='%s'\n",
                       CFStringGetCStringPtr(first, kCFStringEncodingUTF8),
                       CFStringGetCStringPtr(second, kCFStringEncodingUTF8));
@@ -436,7 +420,7 @@ ParseHelperString(const char *str)
 void
 UIMHelperDisconnect()
 {
-#ifdef DEBUG_HELPER
+#if DEBUG_HELPER
   DEBUG_PRINT("UIMHelperDisconnect()\n");
 #endif
 
@@ -454,7 +438,7 @@ UIMHelperDisconnect()
 void
 UIMHelperClose()
 {
-#ifdef DEBUG_HELPER
+#if DEBUG_HELPER
   DEBUG_PRINT("UIMHelperClose()\n");
 #endif
 
@@ -497,18 +481,20 @@ GetPreeditSegment(PreeditSegment *inSegment, UniCharPtr *outStr,
   UniCharPtr tmp = (UniCharPtr) malloc(sizeof(UniChar) * (*outLen + 1));
   memcpy(tmp, *outStr, sizeof(UniChar) * (*outLen));
 
-#ifdef DEBUG_PREEDIT
+#if DEBUG_PREEDIT
   DEBUG_PRINT("GetPreeditSegment() outLen=%lu, fLength=%lu\n",
               *outLen, inSegment->fLength);
 #endif
 
-  (*outStr) = (UniCharPtr) realloc(*outStr, sizeof(UniChar) *
-                                   ((*outLen) + inSegment->fLength + 1));
+  if (inSegment->fLength > 0) {
+    (*outStr) = (UniCharPtr) realloc(*outStr, sizeof(UniChar) *
+                                     ((*outLen) + inSegment->fLength + 1));
 
-  memcpy(*outStr, tmp, sizeof(UniChar) * (*outLen));
-  memcpy(&((*outStr)[(*outLen)]),
-         inSegment->fBuffer, sizeof(UniChar) * inSegment->fLength);
-  *outLen += inSegment->fLength;
+    memcpy(*outStr, tmp, sizeof(UniChar) * (*outLen));
+    memcpy(&((*outStr)[(*outLen)]),
+           inSegment->fBuffer, sizeof(UniChar) * inSegment->fLength);
+    *outLen += inSegment->fLength;
+  }
 
   free(tmp);
 }
@@ -523,7 +509,7 @@ GetPreeditString(MUIMSessionHandle inHandle, UniCharPtr *outStr,
   str = (UniCharPtr) malloc(sizeof(UniChar));
   str[0] = '\0';
 
-#ifdef DEBUG_PREEDIT
+#if DEBUG_PREEDIT
   DEBUG_PRINT("GetPreeditString() fSegmentCount=%lu\n",
               (*inHandle)->fSegmentCount);
 #endif
@@ -533,11 +519,11 @@ GetPreeditString(MUIMSessionHandle inHandle, UniCharPtr *outStr,
     if ((*inHandle)->fSegments[i].fAttr & UPreeditAttr_Cursor) {
       pos = len;
     }
-#ifdef DEBUG_PREEDIT
+#if DEBUG_PREEDIT
     DEBUG_PRINT("GetPreeditString() i=%lu len=%lu\n", i, len);
 #endif
   }
-#ifdef DEBUG_PREEDIT
+#if DEBUG_PREEDIT
   DEBUG_PRINT("GetPreeditString() len=%lu pos=%lu\n", len, pos);
 #endif
 
@@ -552,3 +538,62 @@ GetPreeditString(MUIMSessionHandle inHandle, UniCharPtr *outStr,
   else
     free(str);
 }
+
+void
+SegmentToHiliteRange(MUIMSessionHandle inHandle, TextRangeArrayPtr *inRange)
+{
+  UInt32 i, pos = 0, j = 0;
+
+  (*inRange) = (TextRangeArrayPtr) NewPtrClear(sizeof(short) +
+                                               sizeof(TextRange) *
+                                               ((*inHandle)->fSegmentCount + 1));
+
+  /*
+   * Uim attribute:
+   *
+   * UPreeditAttr_None
+   * UPreeditAttr_UnderLine
+   * UPreeditAttr_Reverse
+   * UPreeditAttr_Cursor
+   * UPreeditAttr_Separator
+   */
+
+  /*
+   * TSM HiliteStyle:
+   *
+   * 1: kCaretPosition
+   * 2: kRawText
+   * 3: kSelectedRawText
+   * 4: kConvertedText
+   * 5: kSelectedConvertedText
+   * 6: kBlockFillText  ?
+   * 7: kOutlineText  ?
+   * 8: kSelectedText
+   */
+
+  for (i = 0; i < (*inHandle)->fSegmentCount; i++) {
+    // cursor
+    // string length is 0 or greaer than 0
+    if ((*inHandle)->fSegments[i].fAttr & UPreeditAttr_Cursor) {
+      (*inRange)->fRange[j].fStart = (*inRange)->fRange[j].fEnd = pos * sizeof(UniChar);
+      (*inRange)->fRange[j].fHiliteStyle = kCaretPosition;
+      j++;
+    }
+
+    if ((*inHandle)->fSegments[i].fLength > 0) {
+      (*inRange)->fRange[j].fStart = pos * sizeof(UniChar);
+      pos += (*inHandle)->fSegments[i].fLength;
+      (*inRange)->fRange[j].fEnd = pos * sizeof(UniChar);
+
+      if ((*inHandle)->fSegments[i].fAttr & UPreeditAttr_Reverse)
+          (*inRange)->fRange[j].fHiliteStyle = kSelectedConvertedText;
+      else if ((*inHandle)->fSegments[i].fAttr & UPreeditAttr_UnderLine)
+        (*inRange)->fRange[j].fHiliteStyle = kConvertedText;
+      else
+        (*inRange)->fRange[j].fHiliteStyle = kRawText;
+      j++;
+    }
+  }
+
+  (*inRange)->fNumOfRanges = j;
+}
